/***
  BEGIN LICENSE
    
  Copyright (C) 2011 Mario Guerriero <mefrio.g@gmail.com>    
  This program is free software: you can redistribute it and/or modify it    
  under the terms of the GNU Lesser General Public License version 3, as published    
  by the Free Software Foundation.
    
  This program is distributed in the hope that it will be useful, but    
  WITHOUT ANY WARRANTY; without even the implied warranties of    
  MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR    
  PURPOSE.  See the GNU General Public License for more details.
    
  You should have received a copy of the GNU General Public License along    
  with this program.  If not, see <http://www.gnu.org/licenses/>    
  
  END LICENSE    
***/ 

namespace Appearance {
    
    public class IOHelper : GLib.Object {

        // Check if the filename has a picture file extension.
        public static bool is_valid_file_type (string fname) {

            // Cache a lowe-cased copy of the file name
            string fname_down = fname.down();
            // Short-circuit if it's not a picture file extension
            return (fname_down.has_suffix(".png") || fname_down.has_suffix(".jpeg") || fname_down.has_suffix(".jpg") || fname_down.has_suffix(".gif"));
        }

        // Quickly count up all of the valid wallpapers in the wallpaper folder.
        public static int count_wallpapers (GLib.File wallpaper_folder) {

            GLib.FileInfo file_info = null;
            int count = 0;
            try {
                // Get an enumerator for all of the plain old files in the wallpaper folder.
                var enumerator = wallpaper_folder.enumerate_children(FileAttribute.STANDARD_NAME + "," + FileAttribute.STANDARD_TYPE, 0);
                // While there's still files left to count
                while ((file_info = enumerator.next_file ()) != null) {
                    // If it's a picture file
                    if (file_info.get_file_type() == GLib.FileType.REGULAR && is_valid_file_type(file_info.get_name())) {
                        count++;
                    }
                }
            } catch(GLib.Error err) {
                // Just for debugging sake
                stdout.printf("Could not pre-scan wallpaper folder. Progress percentage may be off: %s\n", err.message);
            }
            return count;
        }

    }
    
    class WallpaperWidget : Gtk.Grid {
        
        
        public enum ColumnType
        {
          ICON,
          NAME
        }
        
        string WALLPAPER_DIR = "/usr/share/backgrounds";

        GLib.Settings settings;

        private GLib.List<string> dragged = new GLib.List<string> ();

        Gtk.ListStore store;
        GLib.List <Gtk.TreeIter?> iters = new GLib.List <Gtk.TreeIter?> ();
        Gtk.TreeIter selected_plug;
        Gtk.IconView wallpaper_view;
        Gtk.ComboBoxText combo;
        Gtk.ComboBoxText folder_combo;
        Gtk.ColorButton color;
        Gtk.ProgressBar progressbar;
        Gtk.Label progressbar_label;
        string current_wallpaper_path;
        
        public WallpaperWidget () {
            
            settings = new GLib.Settings ("org.gnome.desktop.background");
        
            store = new Gtk.ListStore (2, typeof (Gdk.Pixbuf), typeof (string));
        
            wallpaper_view = new Gtk.IconView ();
            wallpaper_view.set_selection_mode (Gtk.SelectionMode.SINGLE);
            wallpaper_view.set_pixbuf_column (0);
            wallpaper_view.set_model (this.store);
            wallpaper_view.selection_changed.connect (update_wallpaper);
            
            wallpaper_view.drag_data_received.connect (drag_data_received);
            //drag_dest_set (wallpaper_view, DestDefaults.ALL, {e}, Gdk.DragAction.COPY);
            
            /**
             *  Icon view style
             **/
            string icon_style = """
                .wallpaper-view {
                    background-color: #FFFFFF;
                }
                .wallpaper-view:selected {
                    background-color: #9D9D9D;
                    border-color: shade (mix (rgb (34, 255, 120), #fff, 0.5), 0.9);
                }
                .setting-bar {
                    background-color: #FFFFFF;
                }
            """;
            var icon_view_style = new Gtk.CssProvider ();
            try {
                icon_view_style.load_from_data (icon_style, -1);
            } catch (Error e) {
                warning (e.message);
            }
            wallpaper_view.get_style_context ().add_class ("wallpaper-view");
            wallpaper_view.get_style_context ().add_provider (icon_view_style, Gtk.STYLE_PROVIDER_PRIORITY_THEME);
            
            var scrolled = new Gtk.ScrolledWindow (null, null);
            scrolled.set_size_request (250, 300);
            scrolled.set_hexpand(true);
            scrolled.set_vexpand(true);
            scrolled.add (wallpaper_view);
            scrolled.set_margin_left (20);
            scrolled.set_margin_right (20);
            scrolled.set_margin_bottom (5);
            
            
            folder_combo = new Gtk.ComboBoxText ();
            folder_combo.append ("pic", _("Picture folder"));
            folder_combo.append ("sys", _("Default folder"));
            folder_combo.append ("cus", _("Custom folder"));
            folder_combo.changed.connect (update_wallpaper_folder);
            folder_combo.set_active (1);
            folder_combo.set_margin_left (20);
            
            combo = new Gtk.ComboBoxText ();
            combo.append ("none", _("None"));
            combo.append ("wallpaper", _("Whole"));
            combo.append ("centered", _("Centered"));
            combo.append ("scaled", _("Scaled"));
            combo.append ("stretched", _("Stretched"));
            combo.append ("zoom", _("Zoom"));
            combo.changed.connect (update_mode);

            color = new Gtk.ColorButton ();
            color.color_set.connect (update_color);
            color.set_margin_right (20);
            
            var progressbar_grid = new Gtk.Grid ();
            progressbar = new Gtk.ProgressBar ();
            progressbar.set_hexpand(true);
            progressbar_label = new Gtk.Label (_("Importing wallpapers from the default folder"));
            progressbar_label.set_hexpand (true);
            progressbar_label.set_visible(true);
            
            progressbar_grid.attach (progressbar_label, 0, 0, 1, 1);
            progressbar_grid.attach (progressbar, 0, 1, 1, 1);

            load_settings ();
            
            get_style_context ().add_provider (icon_view_style, Gtk.STYLE_PROVIDER_PRIORITY_THEME);
            get_style_context ().add_class ("setting-bar");
            
            var separator = new Gtk.Separator (Gtk.Orientation.HORIZONTAL);
            separator.set_margin_bottom (5);
            attach (scrolled, 0, 0, 4, 1);
            attach (separator, 0, 1, 4, 1);
            attach (folder_combo, 0, 2, 1, 1);
            attach (progressbar_grid, 1, 2, 1, 1);
            attach (combo, 2, 2, 1, 1);
            attach (color, 3, 2, 1, 1);
            set_column_spacing (5);
            set_margin_bottom (5);
        }
    
        void load_settings () {
            combo.set_active_id (settings.get_string ("picture-options"));
            
            Gdk.Color c;
            Gdk.Color.parse (settings.get_string ("primary-color"), out c);
            color.set_color (c);
            
            current_wallpaper_path = settings.get_string ("picture-uri");
        }
    
        void update_wallpaper () {
            var selected = wallpaper_view.get_selected_items ();
            if (selected.length() == 1) {
                GLib.Value filename;
                // Get the filename of the selected wallpaper.
                var item = selected.nth_data(0);
                this.store.get_iter(out this.selected_plug, item);
                this.store.get_value(this.selected_plug, 1, out filename);
                
                foreach (string f in dragged) {
                    if (f == filename.get_string()) current_wallpaper_path = filename.get_string();
                    else current_wallpaper_path = WALLPAPER_DIR + "/" + filename.get_string();
                }

                settings.set_string ("picture-uri", "file://" + filename.get_string ());

            }
        }
        
        void update_color () {
            Gdk.Color c;
            color.get_color (out c);
            settings.set_string ("primary-color", c.to_string ());
        }
        
        void update_mode () {        
            settings.set_string ("picture-options", combo.get_active_id ());
        }
        
        void update_wallpaper_folder () {
            if (folder_combo.get_active () == 0) {
                clean_wallpapers ();
                WALLPAPER_DIR = GLib.Environment.get_user_special_dir (GLib.UserDirectory.PICTURES);
                load_wallpapers ();
            }
            else if (folder_combo.get_active () == 1) {
                clean_wallpapers ();
                WALLPAPER_DIR = "/usr/share/backgrounds";
                load_wallpapers ();
            }
            else if (folder_combo.get_active () == 2) {
                
                var dialog = new Gtk.FileChooserDialog (_("Select a folder"), null, Gtk.FileChooserAction.SELECT_FOLDER);
                dialog.add_button (Gtk.Stock.CANCEL, Gtk.ResponseType.CANCEL);
                dialog.add_button (Gtk.Stock.OPEN, Gtk.ResponseType.ACCEPT);
                dialog.set_default_response (Gtk.ResponseType.ACCEPT);
                
                if (dialog.run () == Gtk.ResponseType.ACCEPT) {
                    clean_wallpapers ();
                    WALLPAPER_DIR = dialog.get_filename ();
                    load_wallpapers ();
                    dialog.destroy ();
                } else dialog.destroy ();

            }
        }
        
        async void load_wallpapers () {
            
            folder_combo.set_sensitive (false);
            
            current_wallpaper_path = settings.get_string ("picture-uri");
            
            // Make the progress bar visible, since we're gonna be using it.
            progressbar.set_visible(true);
            if (WALLPAPER_DIR == "/usr/share/backgrounds")
                progressbar_label.set_text(_("Importing wallpapers from the default folder"));
            else if (WALLPAPER_DIR == GLib.Environment.get_user_special_dir (GLib.UserDirectory.PICTURES))
                progressbar_label.set_text(_("Importing wallpapers from the picture folder"));
            else
                progressbar_label.set_text(_("Importing wallpapers from ") + WALLPAPER_DIR);

            var directory = File.new_for_path (WALLPAPER_DIR);
            // The number of wallpapers we've added so far
            double done = 0.0;
            // Count the # of wallpapers
            int count = IOHelper.count_wallpapers(directory);
            // Enumerator object that will let us read through the wallpapers asynchronously
            var e = yield directory.enumerate_children_async (FileAttribute.STANDARD_NAME, 0, Priority.DEFAULT);
            
            while (true) {
                // Grab a batch of 10 wallpapers
                var files = yield e.next_files_async (10, Priority.DEFAULT);
                // Stop the loop if we've run out of wallpapers
                if (files == null) {
                    break;
                }
                // Loop through and add each wallpaper in the batch
                foreach (var info in files) {
                    // We're going to add another wallpaper
                    done++;
                    // Skip the file if it's not a picture
                    if (!IOHelper.is_valid_file_type(info.get_name())) {
                        continue;
                    }
                    string filename = WALLPAPER_DIR + "/" + info.get_name ();
                    // Create a thumbnail of the image and load it into the Gtk.IconView
                    var image = new Gdk.Pixbuf.from_file_at_scale(filename, 110, 80, false);
                    // Add the wallpaper name and thumbnail to the Gtk.IconView
                    Gtk.TreeIter root;
                    this.store.append(out root);
                    this.store.set(root, 0, image, -1);
                    this.store.set(root, 1, filename, -1);
                    
                    // Select the wallpaper if it is the current wallpaper
                    if (("file://"+filename) == current_wallpaper_path) {
                        this.wallpaper_view.select_path (this.store.get_path (root));
                    }
                    
                    this.iters.append (root);
                    // Update the progress bar
                    progressbar.set_fraction(done/count);
                    // Have GTK update the UI even while we're busy
                    // working on file IO.
                    while(Gtk.events_pending ()) {
                        Gtk.main_iteration();
                    }
                }
            }
            // Hide the progress bar since we're done with it.
            progressbar.set_visible(false);
            progressbar_label.set_text ("");
            
            folder_combo.set_sensitive (true);
        }
        
        void clean_wallpapers () {
            store.clear ();
        }
        
        new void drag_data_received (Gtk.Widget widget, Gdk.DragContext ctx, 
            int x, int y, Gtk.SelectionData sel, uint information, uint timestamp) {           
            if (sel.get_length () > 0){
                string uri = sel.get_uris ()[0];
                File file = File.new_for_uri (uri);
                string filename = file.get_path ();
                string display_name = Filename.display_basename (filename);
                string extension = display_name.split(".")[display_name.split(".").length - 1];
                
                if (extension != "jpg" && extension != "png" && extension != "jpeg" && extension != "gif") {
                    Gtk.drag_finish (ctx, false, false, timestamp);
                    return;
                }
                
                dragged.append (filename);
                
                // Create a thumbnail of the image and load it into the Gtk.IconView
                var image = new Gdk.Pixbuf.from_file_at_scale(filename, 110, 80, false);
                // Add the wallpaper name and thumbnail to the Gtk.IconView
                Gtk.TreeIter root;
                this.store.append(out root);
                this.store.set(root, 0, image, -1);
                this.store.set(root, 1, filename, -1);

                Gtk.drag_finish (ctx, true, false, timestamp);
                return;
            }
            Gtk.drag_finish (ctx, false, false, timestamp);
            return;
        }
    
    }
}